from typing import List, Tuple
import numpy as np
import pandas as pd
import pickle
import tensorflow as tf
from utility.embedder import *
from sklearn.preprocessing import StandardScaler
from pandas.api.types import is_string_dtype
from pathlib import Path

class NetworkCategory:
    """
    Used to store fields related to a given category,
    such as its name, count of unique values and the size of each embedding layer
    """
    def __init__(self, alias: str, unique_values: int):
        self.alias = alias
        self.unique_values = unique_values
        self.embedding_size = self.get_embedding_size(unique_values)
        
    def get_embedding_size(self, unique_values: int) -> int:
        """
        Return the embedding size to be used on the Embedding layer
        :param unique_values: the number of unique values in the given category
        :return: the size to be used on the embedding layer
        """
        size = int(min(np.ceil(unique_values / 2), 50))
        if size < 2:
            return 2
        else:
            return size

class NeuralEmbedder:
    """
    A neural embedder that can learn entity embeddings from categorial features
    """
    def __init__(self,
                 df: pd.DataFrame,
                 target_name: str,
                 train_ratio: float = 0.8,
                 network_layers: List[int] = (32, 32),
                 dropout_rate: float = 0,
                 activation_fn: str = "relu",
                 kernel_initializer: str = "glorot_uniform",
                 regularization_factor: float = 0,
                 loss_fn: str ='binary_crossentropy',
                 optimizer_fn: str = 'Adam',
                 metrics: List[str] = ['accuracy'],
                 epochs: int = 10,
                 batch_size: int = 32,
                 verbose: bool = False,
                 model_path: str = 'models'):
        
        self.check_not_empty_dataframe(df)
        self.check_target_name(target_name)
        self.check_train_ratio(train_ratio)
        self.check_epochs(epochs)
        self.check_batch_size(batch_size)
        
        self.df = df
        self.target_name = target_name
        self.train_ratio = train_ratio
        self.epochs = epochs
        self.batch_size = batch_size
        self.network_layers = network_layers
        self.dropout_rate = dropout_rate
        self.activation_fn = activation_fn
        self.kernel_initializer = kernel_initializer
        self.regularization_factor = regularization_factor
        self.loss_fn = loss_fn
        self.optimizer_fn = optimizer_fn
        self.metrics = metrics
        self.verbose = verbose
        self.model_path = model_path
        self.unique_classes = self.df[self.target_name].nunique()
        
        self.embedded_categories = self.__get_categorial_cols(df, target_name)
        self.numerical_categories = self.__get_numerical_cols(df, target_name)
        
        self.DEFAULT_WEIGHTS_FILENAME = 'weights.pkl'
        self.DEFAULT_LABELS_FILENAME = 'labels.pkl'
        self.DEFAULT_SCALER_FILENAME = 'scaler.pkl'
        self.DEFAULT_PATH_VISUALIZATIONS = 'visualizations'
        self.mode = "emb"
        self.model = self.__make_model()
        
    def __get_categorial_cols(self, df: pd.DataFrame, target_name: str) -> List:
        """
        Returns a list of the categories from a given pandas DataFrame, with the exception of the provided target name
        :param df: the DataFrame
        :param target_name: the name of the target column to not be included
        :return: a List of Category with the df columns except the provided one
        """
        cat_list = []
        for category in df:
            if not category == target_name and is_string_dtype(df[category]): # Only encode object columns
                cat_list.append(NetworkCategory(category, df[category].nunique()))
        return cat_list
    
    def __get_numerical_cols(self, df: pd.DataFrame, target_name:str) -> List:
        """
        Generates a list of numerial categories from a dataframe
        """
        num_list = []
        for category in df:
            if not category == target_name and is_numeric_dtype(df[category]): # Only encode numerical columns
                num_list.append(category)
        return num_list
    
    def __make_model(self) -> tf.keras.Model:
        """
        This method is used to generate our Model containing the Embedding layers alongside with the output layers
        :return: a compiled Model object
        """
        if self.mode == "full":
            dense_inputs, dense_outputs = self._make_numerical_layers(self.kernel_initializer,
                                                                      self.activation_fn)
            embedded_inputs, embedded_outputs = self._make_embedding_layers()
            model_inputs = dense_inputs + embedded_inputs
            model_outputs = dense_outputs + embedded_outputs
        else:
            model_inputs, model_outputs = self._make_embedding_layers()
        
        output_model = self._make_hidden_layers(model_outputs,
                                                self.network_layers,
                                                self.dropout_rate,
                                                self.activation_fn,
                                                self.kernel_initializer,
                                                self.regularization_factor)
        output_model = self._make_final_layer(output_model)

        model = tf.keras.Model(inputs=model_inputs, outputs=output_model)
        model = self._compile_model(model, self.loss_fn, self.optimizer_fn, self.metrics)
        return model
    
    def _make_hidden_layers(self,
                        outputs: List[tf.keras.layers.Layer],
                        network_layers: List[int],
                        dropout_rate: int,
                        activation_fn: str,
                        kernel_initializer: str,
                        regularization_factor: float) -> tf.keras.layers.Layer:
        output_model = tf.keras.layers.Concatenate()(outputs)
        for _, layers in enumerate(network_layers):
            if regularization_factor:
                output_model = tf.keras.layers.Dense(layers,
                                                    kernel_initializer=kernel_initializer,
                                                    kernel_regularizer=tf.keras.regularizers.l2(
                                                        regularization_factor),
                                                    bias_regularizer=tf.keras.regularizers.l2(
                                                        regularization_factor))(output_model)
            else:
                output_model = tf.keras.layers.Dense(layers, 
                                                        kernel_initializer=kernel_initializer)(output_model)
            output_model = tf.keras.layers.Activation(activation_fn)(output_model)
        if dropout_rate:
            output_model = tf.keras.layers.AlphaDropout(dropout_rate)(output_model)
        return output_model
    
    def _make_embedding_layers(self) -> Tuple[List[tf.keras.layers.Layer], List[tf.keras.layers.Layer]]:
        """
        This method is used to generate the list of inputs and output layers where our Embedding layers will be placed
        :return: a tuple containing two lists: the first, the input layers; the second, the output layers
        """
        embedding_inputs = []
        embedding_outputs = []

        for category in self.embedded_categories:
            input_category = tf.keras.layers.Input(shape=(1,))
            output_category = tf.keras.layers.Embedding(input_dim=category.unique_values,
                                        output_dim=category.embedding_size,
                                        name=category.alias)(input_category)
            output_category = tf.keras.layers.Reshape(target_shape=(category.embedding_size,))(output_category)

            embedding_inputs.append(input_category)
            embedding_outputs.append(output_category)

        return embedding_inputs, embedding_outputs

    def _make_numerical_layers(self, kernel_initializer: str,
                               activation_fn: str) -> Tuple[List[tf.keras.layers.Layer],
                                                            List[tf.keras.layers.Layer]]:
        numerical_inputs = []
        numerical_outputs = []
        
        for category in self.numerical_categories:
            input_category = tf.keras.layers.Input(shape=(1,))
            output_category = tf.keras.layers.Dense(1, name=category,
                                                    kernel_initializer=kernel_initializer,
                                                    activation=activation_fn)(input_category)
            
            numerical_inputs.append(input_category)
            numerical_outputs.append(output_category)
            
        return numerical_inputs, numerical_outputs

    def _make_final_layer(self, previous_layer: tf.keras.layers.Layer) -> tf.keras.layers.Layer:
        output_model = tf.keras.layers.Dense(1)(previous_layer)
        output_model = tf.keras.layers.Activation('sigmoid')(output_model)
        return output_model
    
    def _compile_model(self, model: tf.keras.Model, loss_fn: str,
                      optimizer_fn: str, metrics: List[str]) -> tf.keras.Model:
        model.compile(loss=loss_fn, optimizer=optimizer_fn, metrics=metrics)
        return model

    def fit(self, X_train: np.ndarray, y_train: np.ndarray,
            X_valid: np.ndarray, y_valid: np.ndarray,
            callbacks = None, class_weight: dict = None) -> tf.keras.callbacks.History:
        """
        This method is used to fit a given training and validation data into our entity embeddings model
        :param X_train: training features
        :param y_train: training targets
        :param X_valid: validation features
        :param y_valid: validation targets
        :return a History object and model
        """
        history = self.model.fit(x=transpose_to_list(X_train),
                                 y=y_train,
                                 validation_data=(transpose_to_list(X_valid), y_valid),
                                 epochs=self.epochs,
                                 batch_size=self.batch_size,
                                 callbacks=callbacks,
                                 class_weight=class_weight,
                                 verbose=self.verbose)
        return history

    def get_model(self):
        return self.model
    
    def set_weights(self, weights):
        self.model.set_weights(weights)

    def save_model(self) -> None:
        self.model.save(self.model_path)

    def get_embedded_weights(self) -> List:
        weights_embeddings = []
        if self.numerical_categories is not None:
            columns_to_skip = self.numerical_categories + [self.target_name]
        else:
            columns_to_skip = [self.target_name]
        for column in get_all_columns_except(self.df, columns_to_skip):
            weights = self._get_weights_from_layer(column)
            weights_embeddings.append(weights)

        return weights_embeddings

    def _get_weights_from_layer(self, layer_name: str) -> List[np.array]:
        return self.model.get_layer(layer_name).get_weights()[0]
    
    def get_weights_path(self) -> Path:
        """
        Used to return the path of the stored weights
        :return: the pah of the stored weights on disk
        """
        return Path.joinpath(self.model_path, self.DEFAULT_WEIGHTS_FILENAME)

    def get_labels_path(self) -> Path:
        """
        Used to return the path of the stored labels
        :return: the pah of the stored labels on disk
        """
        return Path.joinpath(self.model_path, self.DEFAULT_LABELS_FILENAME)
    
    def get_scaler_path(self) -> Path:
        """
        Used to return the path of the stored scaler
        :return: the pah of the stored scaler on disk
        """
        return Path.joinpath(self.model_path, self.DEFAULT_SCALER_FILENAME)

    def get_visualizations_dir(self) -> Path:
        """
        Used to return the path of the stored visualizations
        :return: the pah of the stored visualizations on disk
        """
        return Path.joinpath(self.model_path, self.DEFAULT_PATH_VISUALIZATIONS)
        
    def save_weights(self, weights: List) -> None:
        with open(self.get_weights_path(), 'wb') as f:
            pickle.dump(weights, f, -1)
            
    def save_labels(self, labels: List) -> None:
        with open(self.get_labels_path(), 'wb') as f:
            pickle.dump(labels, f, -1)
     
    def save_scaler(self, scaler: StandardScaler) -> None:
        with open(self.get_scaler_path(), 'wb') as f:
            pickle.dump(scaler, f, -1)
            
    def make_visualizations_from_network(self, extension: str = 'pdf') -> List[Figure]:
        with open(self.get_labels_path(), 'rb') as f:
            labels = pickle.load(f)

        with open(self.get_weights_path(), 'rb') as f:
            embeddings = pickle.load(f)
            
        n_numerical_cols = len(self.numerical_categories)
        return make_visualizations(labels, embeddings, self.df,
                                   self.get_visualizations_dir(),
                                   extension, n_numerical_cols)
            
    def check_not_empty_dataframe(self, df: pd.DataFrame) -> None:
        if df.empty:
            raise ValueError("You should provide a non-empty pandas dataframe")

    def check_target_name(self, target_name: str) -> None:
        if not target_name:
            raise ValueError("You should provide a non-empty target name")

    def check_target_existent_in_df(self, target_name: str, df: pd.DataFrame) -> None:
        if target_name not in df.columns:
            raise ValueError("You should provide a target variable that is existent on the dataframe")

    def check_train_ratio(self, train_ratio: float) -> None:
        if train_ratio <= 0 or train_ratio >= 1:
            raise ValueError("You should provide a train ratio greater than 0 and smaller than 1")

    def check_epochs(self, epochs: int) -> None:
        if epochs <= 0:
            raise ValueError("You should provide a epoch greater than zero")

    def check_batch_size(self, batch_size: int) -> None:
        if batch_size <= 0:
            raise ValueError("You should provide a batch size greater than zero")