import numpy as np
import pandas as pd
import paths as pt
from sklearn.model_selection import GridSearchCV
from xgboost import XGBClassifier
from tools import data_loader, file_writer
from sklearn.model_selection import StratifiedKFold
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import roc_auc_score, accuracy_score
from sklearn.model_selection import cross_validate, cross_val_score
import matplotlib.pyplot as plt
from pathlib import Path
import paths as pt
import yaml

def main():
    with open(Path.joinpath(pt.CONFIGS_DIR, "settings.yaml"), 'r') as stream:
        settings = yaml.safe_load(stream)
    
    dl = data_loader.FallDataLoader("fall_emb.csv", settings).load_data()
    X, y = dl.get_data()
            
    parameters = { 
        'n_estimators': [400, 800],
        'class_weight': ['balanced'],
        'max_features': ['auto'],
        'max_depth' : [6],
        'min_samples_split' : [10],
        'min_samples_leaf': [3],
            'criterion' : ['gini']
    }
    
    model = RandomForestClassifier(random_state=0,
                                   class_weight="balanced")
    
    rounds = 5
    outer_scores = np.zeros(rounds)
    nested_scores = np.zeros(rounds)
    metric = 'average_precision'
    best_params = list()
    
    for i in range(rounds):
        #Define both cross-validation objects (inner & outer)
        inner_cv = StratifiedKFold(n_splits=5, shuffle=True, random_state=i)
        outer_cv = StratifiedKFold(n_splits=5, shuffle=True, random_state=i)

        # Non-nested parameter search and scoring
        clf = GridSearchCV(estimator=model, param_grid=parameters,
                           cv=inner_cv, scoring=metric)
        clf.fit(X, y)
        outer_scores[i] = clf.best_score_
        best_params.append(clf.best_params_)

        # Nested CV with parameter optimization
        nested_score = cross_val_score(clf, X=X, y=y, cv=outer_cv,
                                       scoring=metric)
        nested_scores[i] = nested_score.mean()
        print(f'Completed round {i+1}, PR AUC: %.3f (%.3f)'
              % (nested_score.mean(), nested_score.std()))
    
    score_difference = outer_scores - nested_scores
    print("Avg. difference of {:6f} with std. dev. of {:6f}."
          .format(score_difference.mean(), score_difference.std()))
    
    # Print the best params per round and score
    for i, best_param in enumerate(best_params):
        print(f"Round {i+1}: {best_param} with score {nested_scores[i]}")
    
    # Plot scores on each round for nested and non-nested cross-validation
    plt.style.use('seaborn')
    plt.tight_layout()
    plt.figure(figsize=(10,5))
    outer_scores_line, = plt.plot(outer_scores, color='orange')
    nested_line, = plt.plot(nested_scores, color='steelblue')
    plt.ylabel("Score", fontsize="14")
    plt.legend([outer_scores_line, nested_line],
            ["Non-Nested CV", "Nested CV"],
            bbox_to_anchor=(0, .4, .5, 0))
    plt.title("Non-Nested vs Nested Cross-Validation",
            x=.5, y=1.1, fontsize="15")
    
    # Plot bar chart of the difference.
    plt.figure(figsize=(10,5))
    plt.tight_layout()
    difference_plot = plt.bar(range(rounds), score_difference)
    plt.xlabel("Individual Trial #")
    plt.legend([difference_plot],
            ["Non-Nested CV - Nested CV Score"],
            bbox_to_anchor=(0, 1, .8, 0))
    plt.ylabel("score difference", fontsize="14")
    plt.show()
            
if __name__ == "__main__":
    main()